#include <iostream>
#include <fstream>
#include <string>
#include <unordered_map>
#include <sstream>

//#include "LabelMap.h"
#include "Graph.h"

#ifndef FILEMANAGER_H_
#define FILEMANAGER_H_

namespace rimlib{

class FileManager {
	static constexpr char * ws = " \t\n\r\f\v";
	// trim from end of string (right)
	static inline std::string& rtrim(std::string& s, const char* t = ws)
	{
	    s.erase(s.find_last_not_of(t) + 1);
	    return s;
	}
	// trim from beginning of string (left)
	static inline std::string& ltrim(std::string& s, const char* t = ws)
	{
	    s.erase(0, s.find_first_not_of(t));
	    return s;
	}
	// trim from both ends of string (left & right)
	static inline std::string& trim(std::string& s, const char* t = ws)
	{
	    return ltrim(rtrim(s, t), t);
	}


public:

	std::unordered_map<std::string, int> nodelabelmap;
	std::unordered_map<std::string, int> edgelabelmap;
	int labelcounter;

	FileManager()
	{
		labelcounter = 0;
	}


	Graph * readGraph( std::string ifile, bool directed)
	{
		Graph * g = NULL;


		std::ifstream myfile (ifile);
		if (myfile.is_open())
		{
			std::string line,s, ll;
			int nof_nodes;
			//std::unordered_map<std::string, int> nodelabelmap;
			//std::unordered_map<std::string, int> edgelabelmap;
			int nof_edges;
			int src, dst, l;

			myfile >> line; //graph name
			myfile >> nof_nodes;

			g = new Graph(directed, nof_nodes);

			for(int i=0; i<nof_nodes; i++)
			{
				myfile >> line;
				std::stringstream ss(line);
				std::string labString;
				std::unordered_set<int> labelsNode;
				while (std::getline(ss, labString, ',')) 
				{
					//label = nodelabelmap.insert({line, nodelabelmap.size()}).second;
					nodelabelmap.insert({labString, nodelabelmap.size()});
					labelsNode.insert(nodelabelmap[labString]);
				}
				g->addNode(i,labelsNode);
			}
			labelcounter =  nodelabelmap.size() + edgelabelmap.size();

			myfile >> nof_edges;
			std::getline(myfile,line);

			for(int i=0; i<nof_edges; i++)
			{
				std::unordered_set< int > edgeLabels;
				std::getline(myfile,line);
				std::stringstream ss(trim(line));
				ss >> src;
				ss >> dst;

//				while(!ss.eof()){
//					ss>>s;
//					std::cout<<"l: "<<s<<std::endl;
//					edgelabelmap.insert({s,edgelabelmap.size()});
//					edgeLabels.insert( edgelabelmap[s] );
//					//ss.ignore(1);
//				}

				ss >> ll;
				std::stringstream ssl(trim(ll));
				while(std::getline(ssl, s, ',')){
					edgelabelmap.insert({s,labelcounter + edgelabelmap.size()});
					edgeLabels.insert( edgelabelmap[s] );

				}

				g->addEdge(src,dst,edgeLabels);
			}
			labelcounter =  nodelabelmap.size() + edgelabelmap.size();

			myfile.close();
		}

		return g;
	}



};

};

#endif
