#include <unordered_set>
//#include <bitset>

#include "sbitset.h"

#ifndef MATCHINGMACHINE_H_
#define MATCHINGMACHINE_H_
namespace rimlib{


enum MamaParentType {PARENTTYPE_IN, PARENTTYPE_OUT, PARENTTYPE_NULL};

class MaMaEdge
{
public:
	int source;
	int target;
	std::unordered_set<int> * attrs;

	MaMaEdge(int _source, int _target, std::unordered_set<int> * _attrs)
		: source(_source), target(_target), attrs(_attrs)
	{
	}
	MaMaEdge(){
		source = -1;
		target = -1;
		attrs = NULL;
	}
};


class MatchingMachine{
public:

	enum NodeFlag {NS_CORE, NS_CNEIGH, NS_UNV};
	int nof_sn;
	int * edges_sizes;				//indexed by state_id
	MaMaEdge ** edges;				//indexed by state_id, map on states  (0,1) = (state0, state1)
	int * map_node_to_state;			//indexed by node_id
	int * map_state_to_node;			//indexed by state_id
	int * parent_state;				//indexed by state_id
	MamaParentType * parent_type;	//indexed by state id

	//public MatchingMachine(Graph g, BitSet[] domains)
	MatchingMachine(Graph &g, sbitset * domains)
	{
//		TIntObjectHashMap[] graphOutAdjList=g.getOutAdjList();
//		nof_sn = graphOutAdjList.length;
//		edges_sizes = new int[nof_sn];
//		edges = new MaMaEdge[nof_sn][];
//		map_node_to_state = new int[nof_sn];
//		map_state_to_node = new int[nof_sn];
//		parent_state = new int[nof_sn];
//		parent_type = new MamaParentType[nof_sn];

		//build(g, domains);

		nof_sn = g.numNodes;
		edges_sizes = (int*) calloc(nof_sn, sizeof(int));
		edges = new MaMaEdge*[nof_sn];
		map_node_to_state = (int*) calloc(nof_sn, sizeof(int));
		map_state_to_node = (int*) calloc(nof_sn, sizeof(int));
		parent_state = (int*) calloc(nof_sn, sizeof(int));
		parent_type = new MamaParentType[nof_sn];

		build(g, domains);
	}

	~MatchingMachine()
	{
		for(int i=0; i<nof_sn; i++){
			delete[] edges[i];
		}
		delete[] edges;
		free(edges_sizes);
		free(map_node_to_state);
		free(map_state_to_node);
		free(parent_state);
		delete[] parent_type;
	}

	void print(){
		std::cout<<"MAMA"<<std::endl;
		for(int i=0; i<nof_sn; i++){
			std::cout<< "n["<<i<<"]->s["<<map_node_to_state[i]<<"]" <<std::endl;
		}
		for(int i=0; i<nof_sn; i++){
			std::cout<<"s["<<i<<"] s2n("<<map_state_to_node[i]<<") ps("<<parent_state[i]<<") pt("<<parent_type[i]<<") ed["<<edges_sizes[i]<<"]"<<std::endl;
			std::cout<<"\t";
			if(edges_sizes[i]==0)
				std::cout<<"no edge";
			for(int j=0; j<edges_sizes[i]; j++){
				std::cout<<"("<<edges[i][j].source<<","<<edges[i][j].target<<")";
			}
			std::cout<<std::endl;
		}
	}

	void build(Graph &g, sbitset * domains)
	{
		int i, j;
		//TIntObjectHashMap[] outAdiacs=g.getOutAdjList();
		//TIntObjectHashMap[] inAdiacs=g.getInAdjList();
		std::unordered_map<int, std::unordered_set<int> > * outAdiacs = g.outAdjList;
		std::unordered_map<int, std::unordered_set<int> > * inAdiacs = g.inAdjList;
		NodeFlag* node_flags = new NodeFlag[nof_sn]; 						//indexed by node_id
		int** weights = new int*[nof_sn];									//indexed by node_id
		int* t_parent_node = (int*) calloc(nof_sn, sizeof(int));			//indexed by node_id
		MamaParentType* t_parent_type = new MamaParentType[nof_sn];		//indexed by node id

		int  nof_single_domains = 0;

		for(int i=0; i<nof_sn; i++){
			node_flags[i] = NS_UNV;
			weights[i] = new int[3];
			weights[i][0] = 0;
			weights[i][1] = 0;
			weights[i][2] = outAdiacs[i].size() + inAdiacs[i].size();
			t_parent_node[i] = -1;
			t_parent_type[i] = PARENTTYPE_NULL;

			//TODO
			//if(domains_size[i] == 1)
			//	nof_single_domains++;
		}

		int si = 0;
		int n;
		int nIT; int ni;
		int nnIT; int nni;
		int nqueueL = 0, nqueueR = 0;
		int maxi, maxv;
		int tmp;

		//TODO
		/*if(nof_single_domains != 0)
		{
		insert single domain
		}*/



		while(si < nof_sn)
		{
			if(nqueueL == nqueueR)
			{
				//if queue is empty....
				maxi = -1;
				maxv = -1;
				nIT = 0;
				while(nIT < nof_sn)
				{
					if(node_flags[nIT]==NodeFlag::NS_UNV &&  weights[nIT][2] > maxv)
					{
						maxv = weights[nIT][2];
						maxi = nIT;
					}
					nIT++;
				}
				map_state_to_node[si] = maxi;
				map_node_to_state[maxi] = si;
				t_parent_type[maxi] = MamaParentType::PARENTTYPE_NULL;
				t_parent_node[maxi] = -1;
				nqueueR++;
				n = maxi;

				/*TIntIterator it=outAdiacs[n].keySet().iterator();
				while(it.hasNext())
				{
					ni=it.next();
					if(ni != n)
						weights[ni][1]++;
				}
				it=inAdiacs[n].keySet().iterator();
				while(it.hasNext())
				{
					ni=it.next();
					if(ni != n)
						weights[ni][1]++;
				}*/

				//for(std::unordered_map<int, std::unordered_set<int> >::iterator it = outAdiacs[n].begin(); it != outAdiacs[n].end(); it++){
				for(auto &it : outAdiacs[n]){
					if( it.first != n){
						weights[it.first][1]++;
					}
				}
				//for(std::unordered_map<int, std::unordered_set<int> >::iterator it = inAdiacs[n].begin(); it != inAdiacs[n].end(); it++){
				for(auto &it : inAdiacs[n]){
					if( it.first != n){
						weights[it.first][1]++;
					}
				}

			}
			if(nqueueL != nqueueR-1)
			{
				maxi = nqueueL;
				for(int mi=maxi+1; mi<nqueueR; mi++)
				{
					if(wcompare(map_state_to_node[mi], map_state_to_node[maxi], weights) < 0)
						maxi = mi;
				}
				tmp = map_state_to_node[nqueueL];
				map_state_to_node[nqueueL] = map_state_to_node[maxi];
				map_state_to_node[maxi] = tmp;
			}
			n = map_state_to_node[si];
			map_node_to_state[n] = si;
			//move queue left limit
			nqueueL++;
			//update nodes' flags & weights
			node_flags[n] = NodeFlag::NS_CORE;
			//TIntIterator it=outAdiacs[n].keySet().iterator();
			//while(it.hasNext())
			//for(std::unordered_map<int, std::unordered_set<int> >::iterator it = outAdiacs[n].begin(); it != outAdiacs[n].end(); it++)
			for(auto &it : outAdiacs[n])
			{
				//ni=it.next();
				ni = it.first;
				if(ni != n)
				{
					weights[ni][0]++;
					weights[ni][1]--;
					if(node_flags[ni] == NodeFlag::NS_UNV)
					{
						node_flags[ni] = NodeFlag::NS_CNEIGH;
						t_parent_node[ni] = n;
//						if(nIT < ssg.out_adj_sizes[n])
							t_parent_type[ni] = MamaParentType::PARENTTYPE_OUT;
//						else
//							t_parent_type[ni] = PARENTTYPE_IN;
						//add to queue
						map_state_to_node[nqueueR] = ni;
						map_node_to_state[ni] = nqueueR;
						nqueueR++;

//						TIntIterator it2=outAdiacs[ni].keySet().iterator();
//						while(it2.hasNext())
//						{
//							nni=it2.next();
//							weights[nni][1]++;
//						}
						//for(std::unordered_map<int, std::unordered_set<int> >::iterator it2 = outAdiacs[ni].begin(); it2 != outAdiacs[ni].end(); it2++)
						for(auto &it2 : outAdiacs[ni])
						{
							weights[it2.first][1]++;
						}
					}
				}
			}
//			it=inAdiacs[n].keySet().iterator();
//			while(it.hasNext())
//			{
//				ni=it.next();
			//for(std::unordered_map<int, std::unordered_set<int> >::iterator it = inAdiacs[n].begin(); it != inAdiacs[n].end(); it++)
			for(auto &it : inAdiacs[n])
			{
				//ni=it.next();
				ni = it.first;
				if(ni != n)
				{
					weights[ni][0]++;
					weights[ni][1]--;
					if(node_flags[ni] == NodeFlag::NS_UNV)
					{
						node_flags[ni] = NodeFlag::NS_CNEIGH;
						t_parent_node[ni] = n;
//						if(nIT < ssg.out_adj_sizes[n])
//							t_parent_type[ni] = PARENTTYPE_OUT;
//						else
							t_parent_type[ni] = MamaParentType::PARENTTYPE_IN;
						//add to queue
						map_state_to_node[nqueueR] = ni;
						map_node_to_state[ni] = nqueueR;
						nqueueR++;

//						TIntIterator it2=inAdiacs[n].keySet().iterator();
//						while(it2.hasNext())
//						{
//							nni=it2.next();
//							weights[nni][1]++;
//						}

						//for(std::unordered_map<int, std::unordered_set<int> >::iterator it2 = inAdiacs[ni].begin(); it2 != inAdiacs[ni].end(); it2++)
						for(auto &it2 : inAdiacs[ni])
						{
							weights[it2.first][1]++;
						}
					}
				}
			}
			si++;
		}




		int e_count,o_e_count,i_e_count;
		for(si = 0; si<nof_sn; si++)
		{
			n = map_state_to_node[si];
			if(t_parent_node[n] != -1)
				parent_state[si] = map_node_to_state[t_parent_node[n]];
			else
				parent_state[si] = -1;
			parent_type[si] = t_parent_type[n];
			e_count = 0;
			o_e_count = 0;
//			TIntIterator it=outAdiacs[n].keySet().iterator();
//			while(it.hasNext())
//			{
//				int idOut=it.next();
			//for(std::unordered_map<int, std::unordered_set<int> >::iterator it = outAdiacs[n].begin(); it != outAdiacs[n].end(); it++)
			for(auto &it : outAdiacs[n])
			{
				//ni=it.next();
				int idOut = it.first;
				if(map_node_to_state[idOut]<si)
				{
					e_count++;
					o_e_count++;
				}
			}
			i_e_count = 0;
//			it=inAdiacs[n].keySet().iterator();
//			while(it.hasNext())
//			{
//				int idIn=it.next();
			//for(std::unordered_map<int, std::unordered_set<int> >::iterator it = inAdiacs[n].begin(); it != inAdiacs[n].end(); it++)
			for(auto &it : inAdiacs[n])
			{
				//ni=it.next();
				int idIn = it.first;
				if(map_node_to_state[idIn]<si)
				{
					e_count++;
					i_e_count++;
				}
			}
			edges_sizes[si] = e_count;
			edges[si] = new MaMaEdge[e_count];
			e_count = 0;
//			it=outAdiacs[n].keySet().iterator();
//			while(it.hasNext())
//			{
//				int idOut=it.next();
			for(auto &it : outAdiacs[n])
			{
				int idOut=it.first;
				if(map_node_to_state[idOut] < si)
				{
					//TIntHashSet setCols=(TIntHashSet) outAdiacs[n].get(idOut);
					//edges[si][e_count]=new MaMaEdge(map_node_to_state[n],map_node_to_state[idOut],setCols);

					edges[si][e_count].source = map_node_to_state[n];
					edges[si][e_count].target = map_node_to_state[idOut];
					edges[si][e_count].attrs = &(outAdiacs[n][idOut]);

					e_count++;
				}
			}
//			it=inAdiacs[n].keySet().iterator();
//			while(it.hasNext())
//			{
//				int idIn=it.next();
			for(auto &it : inAdiacs[n])
			{
				int idIn=it.first;
				if(map_node_to_state[idIn] < si)
				{
					//TIntHashSet setCols=(TIntHashSet) outAdiacs[n].get(idIn);
					//edges[si][e_count]=new MaMaEdge(map_node_to_state[idIn],map_node_to_state[n],setCols);

					edges[si][e_count].source = map_node_to_state[n];
					edges[si][e_count].target = map_node_to_state[idIn];
					edges[si][e_count].attrs = &(outAdiacs[n][idIn]);

					e_count++;
				}
			}
		}
	}


	int wcompare(int i, int j, int ** weights){
		for(int w=0; w<3; w++){
			if(weights[i][w] != weights[j][w]){
				return weights[j][w] - weights[i][w];
			}
		}
		return i-j;
	}

};


}
#endif //MATCHINGMACHINE_H_
