#include <vector>
#include <algorithm>

#include "Graph.h"
#include "MatchingMachine.h"
#include "MatchListener.h"

#ifndef RISOLVER_H_
#define RISOLVER_H_
namespace rimlib{

typedef std::unordered_multiset< std::unordered_set<int>, ushash > powerset_t;

class RISolver{
public:

	MatchingMachine &mama;
	Graph &targetGraph;
	Graph &queryGraph;
	MatchListener &mlistener;


	RISolver( MatchingMachine &_mama, Graph &_targetGraph, Graph &_queryGraph, MatchListener &_mlistener)
		: mama(_mama), targetGraph(_targetGraph), queryGraph(_queryGraph), mlistener(_mlistener)
	{
	}

	void match()
	{
		mlistener.matchcount = 0;
		
		sbitset * domains = computeDomains();
		
		std::vector<int> * symmCond=queryGraph.getSymmetryConditions();
		int nof_sn = queryGraph.numNodes;
		
		std::unordered_map<int, std::unordered_set<int> > * targetOutAdjLists = targetGraph.outAdjList;
		std::unordered_map<int, std::unordered_set<int> > * targetInAdjLists = targetGraph.inAdjList;
		int nofTargetNodes = targetGraph.numNodes;
		int i, j;
		int * parent_state = mama.parent_state;
		MamaParentType * parent_type = mama.parent_type;
		
		int * domain_size = new int[nof_sn];
		for(i=0; i<nof_sn; i++)
		{
			int dsize = 0;
			dsize = 0;
			for(sbitset::iterator IT = domains[i].first_ones(); IT!=domains[i].end(); IT.next_ones())
			{
				dsize++;
			}
			domain_size[i] = dsize;
		}
		
		int ** targetOutAdjArray = new int*[nofTargetNodes];
		int ** targetInAdjArray = new int*[nofTargetNodes];
		int * targetOutAdjArray_size = new int[nofTargetNodes];
		int * targetInAdjArray_size = new int[nofTargetNodes];
		for(i=0; i<nofTargetNodes; i++)
		{
			targetOutAdjArray[i] = new int[ targetOutAdjLists[i].size() ];
			targetOutAdjArray_size[i] = targetOutAdjLists[i].size();
			j = 0;
			for(auto &kv : targetOutAdjLists[i]) {
				targetOutAdjArray[i][j] = kv.first;
				j++;
			}
			targetInAdjArray[i] = new int[ targetInAdjLists[i].size() ];
			targetInAdjArray_size[i] = targetInAdjLists[i].size();
			j = 0;
			for(auto &kv : targetInAdjLists[i]) {
				targetInAdjArray[i][j] = kv.first;
				j++;
			}
		}
		
		int * candidatesIT = new int[nof_sn];
		int * candidatesSize = new int[nof_sn];
		int ** candidates = new int*[nof_sn];
		int * solution = new int[nof_sn];
		for(i=0;i<nof_sn;i++)
			solution[i] = -1;
		std::unordered_set<int> * cmatched = new std::unordered_set<int>[nof_sn];
		bool * matched = (bool*) calloc(nofTargetNodes, sizeof(bool));
		
		for(i=0; i<nof_sn; i++)
		{
			if(parent_type[i] == MamaParentType::PARENTTYPE_NULL)
			{
				int n = mama.map_state_to_node[i];
				candidates[i]=new int[domain_size[n]];
				int k = 0;
				for(sbitset::iterator IT = domains[n].first_ones(); IT!=domains[n].end(); IT.next_ones())
				{
					candidates[i][k]= IT.first;
					k++;
				}

				candidatesSize[i] = domain_size[n];
				candidatesIT[i] = -1;
			}
		}

//#define DEBUG
		int psi = -1;
		int si = 0;
		int ci;
		int sip1;
		while(si != -1)
		{

#ifdef DEBUG
			std::cout<<"----------------------------------------"<<std::endl;
			std::cout<<"state index["<<si<<"]"<<std::endl;
#endif

			if(psi >= si)
			{
				matched[solution[si]] = false;
			}
			ci = -1;
			candidatesIT[si]++;
			while(candidatesIT[si] < candidatesSize[si])
			{
				ci = candidates[si][candidatesIT[si]];

				solution[si] = ci;

#ifdef DEBUG
				std::cout<<"\ttemporary candidate id-"<<ci<<": ";
				std::cout<< !(matched[ci])<<", ";
				std::cout<< domains[mama.map_state_to_node[si]].get(ci)<<", ";
				std::cout<< edgesCheck(si, solution);
				std::cout<< std::endl;
#endif

				if(!matched[ci]
					&& domains[mama.map_state_to_node[si]].get(ci)
					&& condCheck(si, solution, symmCond)
					&& edgesCheck(si, solution)
							){
					break;
				}
				else{
					ci = -1;
				}
				candidatesIT[si]++;
			}

#ifdef DEBUG
			std::cout<<"\tchosen candidate id-"<<ci<<std::endl;
#endif

			if(ci == -1)
			{
				psi = si;
				si--;
			}
			else
			{
				if(si == nof_sn -1)
				{
					mlistener.match(nof_sn, mama.map_state_to_node, solution);
					psi = si;
				}
				else
				{
					matched[solution[si]] = true;
					sip1 = si+1;
					if(parent_type[sip1] != PARENTTYPE_NULL)
					{
						if(parent_type[sip1] == PARENTTYPE_IN)
						{
							candidates[sip1] = targetInAdjArray[solution[parent_state[sip1]]];
							candidatesSize[sip1] = targetInAdjArray_size[solution[parent_state[sip1]]];
						}
						else
						{
							candidates[sip1] = targetOutAdjArray[solution[parent_state[sip1]]];
							candidatesSize[sip1] = targetOutAdjArray_size[solution[parent_state[sip1]]];
						}
					}
					candidatesIT[si +1] = -1;

					psi = si;
					si++;
				}
			}
		}
	}


	bool contains(const powerset_t &large, const powerset_t &small){
		for(auto &k : small){
			if( large.find(k) == large.end() )
				return false;
		}
		return true;
	}

	/*sbitset * computeDomains()
	{
		int numQueryNodes= queryGraph.numNodes;
		int numTargetNodes = targetGraph.numNodes;
		sbitset * domains = new sbitset[numQueryNodes];
		int i=0;
		
		//Compute features for query nodes
		std::unordered_set<int> * queryNodeLabs = queryGraph.nodeLabels;
		std::unordered_map<int, std::unordered_set<int> > * queryOutAdjLists = queryGraph.outAdjList;
		powerset_t * setFeaturesQuery = new powerset_t[numQueryNodes];
		for(i=0;i<numQueryNodes;i++)
		{
			for(std::unordered_map<int, std::unordered_set<int> > ::iterator it = queryOutAdjLists[i].begin(); it != queryOutAdjLists[i].end(); it++)
			{
				int dest = it->first;
				//std::cout<<i<<"-"<<dest<<"\n";
				std::vector< std::unordered_set<int> > powerSetNodeLabs = getPowerSet(queryNodeLabs[dest]);
				std::unordered_set<int> edgeLabels = it->second;
				std::vector< std::unordered_set<int> > powerSetEdgeLabs = getPowerSet(edgeLabels);
				for(auto &sNode : powerSetNodeLabs)
				{
					for(auto &sEdge : powerSetEdgeLabs)
					{
						std::unordered_set<int> setCols(sEdge);
						for(int j : sNode)
							setCols.insert(j);
						//std::cout<<"-->"; print_unordered_set(setCols);std::cout<<std::endl;
						setFeaturesQuery[i].insert(setCols);
					}
				}
			}
		}
		

		std::unordered_map<int, std::unordered_set<int> > * targetOutAdjLists = targetGraph.outAdjList;
		std::unordered_set<int> * targetNodeLabs = targetGraph.nodeLabels;
		for(i=0;i<numTargetNodes;i++)
		{
			//Check if target node label is compatible with some query node label
			int j=0;
			for(j=0;j<numQueryNodes;j++)
			{
				if(includes(&targetNodeLabs[i], &queryNodeLabs[j]))
					break;
			}
			if(j<numQueryNodes)
			{
				//Compute features for target node i
				powerset_t setFeaturesTarget;
				for(std::unordered_map<int, std::unordered_set<int> > ::iterator it = targetOutAdjLists[i].begin(); it != targetOutAdjLists[i].end(); it++)
				{
					int dest = it->first;
					std::vector< std::unordered_set<int> > powerSetNodeLabs = getPowerSet(targetNodeLabs[dest]);
					std::unordered_set<int> edgeLabels = it->second;
					std::vector< std::unordered_set<int> > powerSetEdgeLabs = getPowerSet(edgeLabels);
					for(auto &sNode : powerSetNodeLabs)
					{
						for(auto &sEdge : powerSetEdgeLabs)
						{
							std::unordered_set<int> setCols(sEdge);
							for(int x : sNode)
								setCols.insert(x);
							//std::cout<<"-->"; print_unordered_set(setCols);std::cout<<std::endl;
							setFeaturesTarget.insert(setCols);
						}
					}
				}
				//Find compatible query nodes and update domains
				for(j=0;j<numQueryNodes;j++)
				{
					if(includes(&targetNodeLabs[i], &queryNodeLabs[j]) && contains(setFeaturesTarget, setFeaturesQuery[j]))
					{
						domains[j].set(i,true);
						//std::cout<<i<<" - "<<j<<"\n";
					}
				}
			}
		}

		//AC consistency
		int ra, qb, rb;
		bool notfound;
		bool changes=true;
		while(changes)
		{
			changes=false;
			for(int qa=0; qa<numQueryNodes; qa++)
			{
				for(sbitset::iterator qaIT=domains[qa].first_ones(); qaIT!=domains[qa].end(); qaIT.next_ones())
				{
					ra = qaIT.first;
					for(auto &it_qb : queryOutAdjLists[qa])
					{
						qb = it_qb.first;
						notfound = true;
						for(auto &it_rb : targetOutAdjLists[ra])
						{
							rb = it_rb.first;
							if(domains[qb].get(rb) 
								//&& std::includes(it_rb.second.begin(), it_rb.second.end(), it_qb.second.begin(), it_qb.second.end())
								&& includes(&it_rb.second, &it_qb.second))
							{
								notfound = false;
								break;
							}
						}
						if(notfound)
						{
							domains[qa].set(ra, false);
							changes=true;
						}
					}
				}
			}
		}
		
		return domains;
	}*/

	sbitset * computeDomains()
	{
		int numQueryNodes= queryGraph.numNodes;
		int numTargetNodes = targetGraph.numNodes;
		sbitset * domains = new sbitset[numQueryNodes];
		int i=0, j=0;
		
		std::unordered_map<int, std::unordered_set<int> > * queryInAdjLists = queryGraph.inAdjList;
		std::unordered_map<int, std::unordered_set<int> > * targetInAdjLists = targetGraph.inAdjList;
		std::unordered_map<int, std::unordered_set<int> > * queryOutAdjLists = queryGraph.outAdjList;
		std::unordered_map<int, std::unordered_set<int> > * targetOutAdjLists = targetGraph.outAdjList;
		std::unordered_set<int> * queryNodeLabs = queryGraph.nodeLabels;
		std::unordered_set<int> * targetNodeLabs = targetGraph.nodeLabels;
		for(i=0;i<numTargetNodes;i++)
		{
			//Find compatible query nodes and update domains
			for(j=0;j<numQueryNodes;j++)
			{
				if(includes(&targetNodeLabs[i], &queryNodeLabs[j]) 
				&& queryOutAdjLists[j].size()<=targetOutAdjLists[i].size()
				&& queryInAdjLists[j].size()<=targetInAdjLists[i].size())
				{
					domains[j].set(i,true);
				}
			}
		}

		//AC consistency
		int ra, qb, rb;
		bool notfound;
		bool changes=true;
		while(changes)
		{
			changes=false;
			for(int qa=0; qa<numQueryNodes; qa++)
			{
				for(sbitset::iterator qaIT=domains[qa].first_ones(); qaIT!=domains[qa].end(); qaIT.next_ones())
				{
					ra = qaIT.first;
					for(auto &it_qb : queryOutAdjLists[qa])
					{
						qb = it_qb.first;
						notfound = true;
						for(auto &it_rb : targetOutAdjLists[ra])
						{
							rb = it_rb.first;
							if(domains[qb].get(rb) 
								//&& std::includes(it_rb.second.begin(), it_rb.second.end(), it_qb.second.begin(), it_qb.second.end())
								&& includes(&it_rb.second, &it_qb.second))
							{
								notfound = false;
								break;
							}
						}
						if(notfound)
						{
							domains[qa].set(ra, false);
							changes=true;
						}
					}
				}
			}
		}
		
		return domains;
	}
	
	bool condCheck(int si, int * solution, std::vector<int> * symmCond)
	{
		bool condCheck=true;
		std::vector<int> &condNode = symmCond[mama.map_state_to_node[si]];
		for(int ii=0;ii<condNode.size();ii++)
		{
			int targetNode=solution[mama.map_node_to_state[condNode.at(ii)]];
			if(solution[si]<targetNode)
			{
				condCheck = false;
				break;
			}
		}
		return condCheck;
	}

	bool includes(std::unordered_set<int> * large, std::unordered_set<int> * small){
		for(int i : *small){
			if(large->find(i) == large->end()){
				return false;
			}
		}
		return true;
	}

	bool edgesCheck(int si, int * solution)
	{
		//TIntObjectHashMap[] targetOutAdjLists=targetGraph.getOutAdjList();
		std::unordered_map<int, std::unordered_set<int> > * targetOutAdjLists = targetGraph.outAdjList;

		for(int me=0; me<mama.edges_sizes[si]; me++)
		{
			int source = solution[mama.edges[si][me].source];
			int target = solution[mama.edges[si][me].target];
			//if(!targetOutAdjLists[source].containsKey(target))
			if(targetOutAdjLists[source].find(target) == targetOutAdjLists[source].end()){
#ifdef DEBUG
				std::cout<<"!("<<source<<","<<target<<")";
#endif
				return false;
			}
			else
			{
				std::unordered_set<int> * setColsQuery = mama.edges[si][me].attrs;
				if(setColsQuery != NULL){
					std::unordered_set<int> * setColsTarget =  &(targetOutAdjLists[source][target]);

#ifdef DEBUG
					std::cout<<"{";
					for(auto &it : *(setColsQuery)){
						std::cout<<std::to_string(it)<<",";
					}
					std::cout<<"}<>{";
					for(auto &it : *(setColsTarget)){
						std::cout<<std::to_string(it)<<",";
					}
					std::cout<<"}";
#endif
					//if( setColsQuery->size()>0 &&
					if(
							//!(std::includes(setColsTarget->begin(), setColsTarget->end(), setColsQuery->begin(), setColsQuery->end()))
							!(includes(setColsTarget, setColsQuery))
					){
#ifdef DEBUG
						std::cout<<"!!("<<source<<","<<target<<")";
#endif
						return false;
					}
				}
#ifdef DEBUG
				else{
					std::cout<<"NoColors";
				}
#endif

//				TIntHashSet setColsQuery=mama.edges[si][me].attr;
//				if(setColsQuery!=null)
//				{
//					TIntHashSet setColsTarget = (TIntHashSet) targetOutAdjLists[source].get(target);
//					if(!setColsQuery.isEmpty() && !setColsTarget.containsAll(setColsQuery))
//						return false;
//				}
			}
		}
		return true;
	}
	
	std::vector< std::unordered_set<int> > getPowerSet(std::unordered_set<int> &setColors)
	{

//		std::cout<<"powerset(";
//		print_unordered_set(setColors);
//		std::cout<<")"<<std::endl;

		std::vector< std::unordered_set<int> > powerSet;
		int n = setColors.size();
		for (int i = 0; i < (1<<n); i++)
		{
			std::unordered_set<int> set;
			std::unordered_set<int>::iterator it = setColors.begin();
			for (int j = 0; j < n; j++)
			{
				int col = *it; it++;
				if ((i & (1 << j)) > 0)
					set.insert(col);
			}
			if(set.size()>0){
//				std::cout<<" ->";print_unordered_set(set);std::cout<<std::endl;
				powerSet.push_back(set);
			}
			/*TIntHashSet set=new TIntHashSet();
			TIntIterator it=setColors.iterator();
			for (int j = 0; j < n; j++)
			{
				int col=it.next();
				if ((i & (1 << j)) > 0)
					set.add(col);
			}
			if(set.size()>0)
				powerSet.add(set);
			*/
		}
		return powerSet;
	}
	
};

}
#endif //RISOLVER_H_
